#!/usr/bin/env python3
"""Index datasets found from an SQS queue into Postgres
"""
import concurrent
import json
import logging
import os
import sys
from typing import Any, Dict, Generator, Optional, Tuple

import click
import pystac
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from odc.apps.dc_tools._stac import stac_transform, stac_transform_absolute
from odc.apps.dc_tools.utils import (
    SkippedException,
    allow_unsafe,
    archive_less_mature,
    bbox,
    index_update_dataset,
    limit,
    statsd_gauge_reporting,
    statsd_setting,
    update_if_exists_flag,
    publish_action,
)
from pystac.item import Item
from pystac_client import Client

logging.basicConfig(
    level=logging.WARNING,
    format="%(asctime)s: %(levelname)s: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S",
)


def _parse_options(options: Optional[str]) -> Dict[str, Any]:
    parsed_options = {}

    if options is not None:
        for option in options.split("#"):
            try:
                key, value = option.split("=")

                try:
                    value = json.loads(value)
                except Exception:  # pylint:disable=broad-except
                    logging.warning(
                        "Failed to handle value %s for key %s as JSON, using str",
                        value,
                        key,
                    )
                parsed_options[key] = value
            except Exception:  # pylint:disable=broad-except
                logging.warning(
                    "Couldn't parse option %s format is key=value",
                    option,
                    exc_info=True,
                )

    return parsed_options


def _guess_location(
    item: pystac.Item, rewrite: Optional[Tuple[str, str]] = None
) -> Tuple[str, bool]:
    self_link = None
    asset_link = None
    relative = True

    for link in item.links:
        if link.rel == "self":
            self_link = link.target

        # Override self with canonical
        if link.rel == "canonical":
            self_link = link.target
            break

    for _, asset in item.assets.items():
        if "geotiff" in asset.media_type:
            asset_link = os.path.dirname(asset.href)
            break

    if rewrite is not None:
        for _, asset in item.assets.items():
            if "geotiff" in asset.media_type:
                asset.href = asset.href.replace(rewrite[0], rewrite[1])

    # If the metadata and the document are not on the same path,
    # we need to use absolute links and not relative ones.
    if (self_link and asset_link) and os.path.dirname(self_link) != os.path.dirname(
        asset_link
    ):
        relative = False

    return self_link, relative


def item_to_meta_uri(
    item: Item,
    rewrite: Optional[Tuple[str, str]] = None,
    rename_product: Optional[str] = None,
) -> Generator[Tuple[dict, str, bool], None, None]:
    uri, relative = _guess_location(item, rewrite)
    metadata = item.to_dict()
    if rename_product is not None:
        metadata["properties"]["odc:product"] = rename_product
    stac = metadata

    if relative:
        metadata = stac_transform(metadata)
    else:
        metadata = stac_transform_absolute(metadata)

    return (metadata, uri, stac)


def process_item(
    item: Item,
    dc: Datacube,
    doc2ds: Doc2Dataset,
    update_if_exists: bool,
    allow_unsafe: bool,
    rewrite: Optional[Tuple[str, str]] = None,
    rename_product: Optional[str] = None,
    archive_less_mature: bool = False,
    publish_action: bool = False,
):
    meta, uri, stac = item_to_meta_uri(item, rewrite, rename_product)
    index_update_dataset(
        meta,
        uri,
        dc,
        doc2ds,
        update_if_exists=update_if_exists,
        allow_unsafe=allow_unsafe,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        stac_doc=stac,
    )


def stac_api_to_odc(
    dc: Datacube,
    update_if_exists: bool,
    config: dict,
    catalog_href: str,
    allow_unsafe: bool = True,
    rewrite: Optional[Tuple[str, str]] = None,
    rename_product: Optional[str] = None,
    archive_less_mature: bool = False,
    publish_action: Optional[str] = None,
) -> Tuple[int, int, int]:
    doc2ds = Doc2Dataset(dc.index)
    client = Client.open(catalog_href)

    search = client.search(**config)
    n_items = search.matched()
    if n_items is not None:
        logging.info("Found %s items to index", n_items)
        if n_items == 0:
            logging.warning("Didn't find any items, finishing.")
            return 0, 0, 0
    else:
        logging.warning("API did not return the number of items.")

    # Do the indexing of all the things
    success = 0
    failure = 0
    skipped = 0

    sys.stdout.write("\rIndexing from STAC API...\n")
    with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
        future_to_item = {
            executor.submit(
                process_item,
                item,
                dc,
                doc2ds,
                update_if_exists=update_if_exists,
                allow_unsafe=allow_unsafe,
                rewrite=rewrite,
                rename_product=rename_product,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
            ): item.id
            for item in search.get_all_items()
        }
        for future in concurrent.futures.as_completed(future_to_item):
            item = future_to_item[future]
            try:
                _ = future.result()
                success += 1
                if success % 10 == 0:
                    sys.stdout.write(f"\rAdded {success} datasets...")
            except SkippedException:
                skipped += 1
            except Exception:  # pylint:disable=broad-except
                logging.exception("Failed to handle item %s", item)
                failure += 1
    sys.stdout.write("\r")

    return success, failure, skipped


@click.command("stac-to-dc")
@limit
@update_if_exists_flag
@allow_unsafe
@click.option(
    "--catalog-href",
    type=str,
    default="https://earth-search.aws.element84.com/v0/",
    help="URL to the catalog to search",
)
@click.option(
    "--collections",
    type=str,
    default=None,
    help="Comma separated list of collections to search",
)
@bbox
@click.option(
    "--datetime",
    type=str,
    default=None,
    help="Dates to search, either one day or an inclusive range, e.g. 2020-01-01 or 2020-01-01/2020-01-02",
)
@click.option(
    "--options",
    type=str,
    default=None,
    help="Other search terms, as a # separated list, i.e., --options=cloud_cover=0,100#sky=green",
)
@click.option(
    "--rewrite-assets",
    type=str,
    default=None,
    help=(
        "Rewrite asset hrefs, for example, to change from "
        "HTTPS to S3 URIs, --rewrite-assets=https://example.com/,s3://"
    ),
)
@click.option(
    "--rename-product",
    type=str,
    default=None,
    help=(
        "Name of product to overwrite collection(s) names, "
        "only one product name can overwrite, despite multiple collections "
    ),
)
@archive_less_mature
@publish_action
@statsd_setting
def cli(
    limit,
    update_if_exists,
    allow_unsafe,
    catalog_href,
    collections,
    bbox,
    datetime,
    options,
    rewrite_assets,
    rename_product,
    statsd_setting,
    archive_less_mature,
    publish_action,
):
    """
    Iterate through STAC items from a STAC API and add them to datacube.
    """
    config = _parse_options(options)
    rewrite = None

    # Format the search terms
    if bbox:
        config["bbox"] = list(map(float, bbox.split(",")))

    if collections:
        config["collections"] = collections.split(",")

    if datetime:
        config["datetime"] = datetime

    # Always set the limit, because some APIs will stop at an arbitrary
    # number if max_items is not None.
    config["max_items"] = limit

    if rewrite_assets is not None:
        rewrite = list(rewrite_assets.split(","))
        if len(rewrite) != 2:
            raise ValueError(
                "Rewrite assets argument needs to be two strings split by ','"
            )

    # Do the thing
    dc = Datacube()
    added, failed, skipped = stac_api_to_odc(
        dc,
        update_if_exists,
        config,
        catalog_href,
        allow_unsafe=allow_unsafe,
        rewrite=rewrite,
        rename_product=rename_product,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
    )

    print(
        f"Added {added} Datasets, failed {failed} Datasets, skipped {skipped} Datasets"
    )
    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:stac_api_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:stac_api_to_dc", "action:failed"], statsd_setting
        )
        statsd_gauge_reporting(
            skipped, ["app:stac_api_to_dc", "action:skipped"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""Build S3 iterators using odc-tools
and index datasets found into RDS
"""
import logging
import sys
from typing import Tuple

import click
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from odc.aio import S3Fetcher, s3_find_glob
from odc.apps.dc_tools._docs import parse_doc_stream
from odc.apps.dc_tools._stac import stac_transform, stac_transform_absolute
from odc.apps.dc_tools.utils import (
    IndexingException,
    SkippedException,
    allow_unsafe,
    archive_less_mature,
    fail_on_missing_lineage,
    index_update_dataset,
    no_sign_request,
    request_payer,
    skip_check,
    skip_lineage,
    statsd_gauge_reporting,
    statsd_setting,
    transform_stac,
    transform_stac_absolute,
    update_flag,
    update_if_exists_flag,
    verify_lineage,
    publish_action,
)

logging.basicConfig(
    level=logging.WARNING,
    format="%(asctime)s: %(levelname)s: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S",
)


def doc_error(uri, doc):
    """Log the internal errors parsing docs"""
    logging.exception("Failed to parse doc at %s", uri)


def dump_to_odc(
    document_stream,
    dc: Datacube,
    products: list,
    transform=None,
    update=False,
    update_if_exists=False,
    allow_unsafe=False,
    archive_less_mature=None,
    publish_action=None,
    **kwargs,
) -> Tuple[int, int, int]:
    doc2ds = Doc2Dataset(dc.index, products=products, **kwargs)

    ds_added = 0
    ds_failed = 0
    ds_skipped = 0
    uris_docs = parse_doc_stream(
        ((doc.url, doc.data) for doc in document_stream),
        on_error=doc_error,
        transform=transform,
    )

    found_docs = False
    for uri, metadata in uris_docs:
        found_docs = True
        try:
            index_update_dataset(
                metadata,
                uri,
                dc,
                doc2ds,
                update=update,
                update_if_exists=update_if_exists,
                allow_unsafe=allow_unsafe,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
            )
            ds_added += 1
        except IndexingException:
            logging.exception("Failed to index dataset %s", uri)
            ds_failed += 1
        except SkippedException:
            ds_skipped += 1
    if not found_docs:
        raise IndexingException("Doc stream was empty")

    return ds_added, ds_failed, ds_skipped


@click.command("s3-to-dc")
@skip_lineage
@fail_on_missing_lineage
@verify_lineage
@transform_stac
@transform_stac_absolute
@update_flag
@update_if_exists_flag
@allow_unsafe
@skip_check
@no_sign_request
@statsd_setting
@request_payer
@archive_less_mature
@publish_action
@click.argument("uri", type=str, nargs=1)
@click.argument("product", type=str, nargs=1)
def cli(
    skip_lineage,
    fail_on_missing_lineage,
    verify_lineage,
    stac,
    absolute,
    update,
    update_if_exists,
    allow_unsafe,
    skip_check,
    no_sign_request,
    statsd_setting,
    request_payer,
    archive_less_mature,
    publish_action,
    uri,
    product,
):
    """Iterate through files in an S3 bucket and add them to datacube"""

    transform = None
    if stac:
        if absolute:
            transform = stac_transform_absolute
        else:
            transform = stac_transform

    candidate_products = product.split()

    opts = {}
    if request_payer:
        opts["RequestPayer"] = "requester"

    # Check datacube connection and products
    dc = Datacube()
    odc_products = dc.list_products().name.values

    odc_products = set(odc_products)
    if not set(candidate_products).issubset(odc_products):
        missing_products = list(set(candidate_products) - odc_products)
        print(
            f"Error: Requested Product/s {', '.join(missing_products)} {'is' if len(missing_products) == 1 else 'are'} "
            "not present in the ODC Database",
            file=sys.stderr,
        )
        sys.exit(1)

    # Get a generator from supplied S3 Uri for candidate documents
    fetcher = S3Fetcher(aws_unsigned=no_sign_request)
    # Grab the URL from the resulting S3 item
    document_stream = (
        url.url for url in s3_find_glob(uri, skip_check=skip_check, s3=fetcher, **opts)
    )

    added, failed, skipped = dump_to_odc(
        fetcher(document_stream),
        dc,
        candidate_products,
        skip_lineage=skip_lineage,
        fail_on_missing_lineage=fail_on_missing_lineage,
        verify_lineage=verify_lineage,
        transform=transform,
        update=update,
        update_if_exists=update_if_exists,
        allow_unsafe=allow_unsafe,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
    )

    print(
        f"Added {added} datasets, skipped {skipped} datasets and failed {failed} datasets."
    )
    if statsd_setting:
        statsd_gauge_reporting(added, ["app:s3_to_dc", "action:added"], statsd_setting)
        statsd_gauge_reporting(
            skipped, ["app:s3_to_dc", "action:skipped"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:s3_to_dc", "action:failed"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()



"""Crawl Thredds for prefixes and fetch YAML's for indexing
and dump them into a Datacube instance
"""
import logging
import sys
from typing import List, Tuple

import click
from datacube import Datacube
from odc.apps.dc_tools.utils import statsd_gauge_reporting, statsd_setting
from odc.thredds import download_yamls, thredds_find_glob

from ._docs import from_yaml_doc_stream


def dump_list_to_odc(
    yaml_content_list: List[Tuple[bytes, str, str]],
    dc: Datacube,
    products: List[str],
    **kwargs,
):
    expand_stream = (
        ("https://" + d[1], d[0]) for d in yaml_content_list if d[0] is not None
    )

    ds_stream = from_yaml_doc_stream(
        expand_stream, dc.index, products=products, **kwargs
    )
    ds_added = 0
    ds_failed = 0
    # Consume chained streams to DB
    for result in ds_stream:
        ds, err = result
        if err is not None:
            logging.error(err)
            ds_failed += 1
        else:
            logging.info(ds)
            # TODO: Potentially wrap this in transactions and batch to DB
            # TODO: Capture UUID's from YAML and perform a bulk has
            try:
                dc.index.datasets.add(ds)
                ds_added += 1
            except Exception as e:
                logging.error(e)
                ds_failed += 1

    return ds_added, ds_failed


@click.command("thredds-to-dc")
@click.option(
    "--skip-lineage",
    is_flag=True,
    default=False,
    help="Default is not to skip lineage. Set to skip lineage altogether.",
)
@click.option(
    "--fail-on-missing-lineage/--auto-add-lineage",
    is_flag=True,
    default=True,
    help=(
        "Default is to fail if lineage documents not present in the database. "
        "Set auto add to try to index lineage documents."
    ),
)
@click.option(
    "--verify-lineage",
    is_flag=True,
    default=False,
    help="Default is no verification. Set to verify parent dataset definitions.",
)
@statsd_setting
@click.argument("uri", type=str, nargs=1)
@click.argument("product", type=str, nargs=1)
def cli(
    skip_lineage: bool,
    fail_on_missing_lineage: bool,
    verify_lineage: bool,
    statsd_setting: str,
    uri: str,
    product: str,
):
    skips = [".*NBAR.*", ".*SUPPLEMENTARY.*", ".*NBART.*", ".*/QA/.*"]
    select = [".*ARD-METADATA.yaml"]
    candidate_products = product.split()
    print(f"Crawling {uri} on Thredds")
    print(f"Matching to {candidate_products}")
    yaml_urls = thredds_find_glob(uri, skips, select)
    print(f"Found {len(yaml_urls)} datasets")

    yaml_contents = download_yamls(yaml_urls)

    # Consume generator and fetch YAML's
    dc = Datacube()
    added, failed = dump_list_to_odc(
        yaml_contents,
        dc,
        candidate_products,
        skip_lineage=skip_lineage,
        fail_on_missing_lineage=fail_on_missing_lineage,
        verify_lineage=verify_lineage,
    )

    print(f"Added {added} Datasets, Failed {failed} Datasets")
    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:thredds_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:thredds_to_dc", "action:failed"], statsd_setting
        )



"""Add or update a list of ODC products. Intended to be used to
systematically maintain a CSV of products and synchronise it with
a database"""

import logging
import sys
from collections import Counter, namedtuple
from csv import DictReader
from typing import Any, Dict, List, Optional, Generator, Tuple

import click
import datacube
import fsspec
import yaml
from datacube import Datacube
from odc.apps.dc_tools.utils import (
    update_if_exists_flag,
    statsd_gauge_reporting,
    statsd_setting,
)

Product = namedtuple("Product", ["name", "doc"])

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s: %(levelname)s: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S",
)


def _get_product(product_path: str) -> List[Dict[str, Any]]:
    """Returns yaml document"""
    try:
        with fsspec.open(product_path, mode="r") as f:
            return list(yaml.safe_load_all(f))
    except Exception:  # pylint:disable=broad-except
        logging.exception("Failed to get document from %s", product_path)
        return []


def _parse_csv(csv_path: str) -> Generator[Product, None, None]:
    """Parses the CSV and returns a dict of name: yaml_file_path"""

    with fsspec.open(csv_path, mode="r") as f:
        reader = DictReader(f)
        for row in reader:
            names = row["product"].split(";")
            content = _get_product(row["definition"])

            # Do some QA
            fail = False
            # Only return value if we find contents in the document
            if content is None:
                fail = True

            # Check we have the same number of names as content
            if len(names) != len(content):
                logging.error(
                    "%s product names and %s documents found. This is different!",
                    len(names),
                    len(content),
                )
                fail = True

            # Check we have the same names as are in the product definitions
            content_names = [d["name"] for d in content]
            if Counter(content_names) != Counter(names):
                logging.error("%s is not the same as %s", names, content_names)
                fail = True

            if fail:
                yield Product(row["product"], None)
                continue

            # There's only one name in names, so yield it
            if len(names) == 1:
                yield Product(names[0], content[0])
            else:
                # Handle multiple documents in a single file
                for doc in content:
                    # Since we checked all the names, we can do this safely
                    yield Product(doc["name"], doc)


def add_update_products(
    dc: Datacube, csv_path: str, update_if_exists: Optional[bool] = False
) -> Tuple[int, int, int]:
    # Parse csv file
    new_products = list(_parse_csv(csv_path))
    logging.info("Found %s products in the CSV %s", len(new_products), csv_path)

    # List existing products
    products = dc.list_products(with_pandas=False)
    existing_names = [product["name"] for product in products]
    logging.info("Found %s products in the Datacube", len(existing_names))

    added, updated, failed = 0, 0, 0

    for product in new_products:
        if product.doc is None:
            failed += 1
            continue
        # Add new products
        try:
            if product.name not in existing_names:
                dc.index.products.add_document(product.doc)
                added += 1
                logging.info("Added product %s", product.name)
            # Update existing products, if required
            elif update_if_exists:
                dc.index.products.update_document(
                    product.doc, allow_unsafe_updates=True
                )
                updated += 1
                logging.info("Updated product %s", product.name)
        except Exception:  # pylint:disable=broad-except
            failed += 1
            logging.exception("Failed to add/update product %s", product.name)

    # Return results
    return added, updated, failed


@click.command("dc-sync-products")
@click.argument("csv-path", nargs=1)
@update_if_exists_flag
@statsd_setting
def cli(csv_path: str, update_if_exists: bool, statsd_setting: str):
    # Check we can connect to the Datacube
    dc = datacube.Datacube(app="add_update_products")
    logging.info(
        "Starting up: connected to Datacube, and update-if-exists is: %s",
        update_if_exists,
    )

    # TODO: Add in some QA/QC checks
    added, updated, failed = add_update_products(dc, csv_path, update_if_exists)

    print(f"Added: {added}, Updated: {updated} and Failed: {failed}")
    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app: add_update_products", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app: add_update_products", "action:failed"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app: add_update_products", "action:updated"], statsd_setting
        )

    # If nothing failed then this exists with success code 0
    sys.exit(failed)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""
Index the Copernicus DEM automatically.
"""
import concurrent.futures
import logging
import sys
from math import ceil, floor
from typing import Tuple

import click
import pystac
import rasterio
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.utils import read_documents
from odc.apps.dc_tools.utils import (
    SkippedException,
    archive_less_mature,
    bbox,
    index_update_dataset,
    limit,
    update_if_exists_flag,
    publish_action,
    statsd_gauge_reporting,
    statsd_setting,
)
from rio_stac import create_stac_item

from ._stac import stac_transform

PRODUCTS = {
    "cop_30": (
        "https://raw.githubusercontent.com/opendatacube/"
        "datacube-dataset-config/master/products/dem_cop_30.odc-product.yaml"
    ),
    "cop_90": (
        "https://raw.githubusercontent.com/opendatacube/"
        "datacube-dataset-config/master/products/dem_cop_90.odc-product.yaml"
    ),
}

# URIs need north/south, which is N00 and east/west, which is E000
URI_TEMPLATES = {
    "cop_30": (
        "https://copernicus-dem-30m.s3.eu-central-1.amazonaws.com/"
        "Copernicus_DSM_COG_10_{ns}_00_{ew}_00_DEM/Copernicus_DSM_COG_10_{ns}_00_{ew}_00_DEM.tif"
    ),
    "cop_90": (
        "https://copernicus-dem-90m.s3.eu-central-1.amazonaws.com/"
        "Copernicus_DSM_COG_30_{ns}_00_{ew}_00_DEM/Copernicus_DSM_COG_30_{ns}_00_{ew}_00_DEM.tif"
    ),
}


def add_cop_dem_product(dc: Datacube, product):
    if product in PRODUCTS:
        product_uri = PRODUCTS[product]
    else:
        raise ValueError(f"Unknown product {product}")

    for _, doc in read_documents(product_uri):
        dc.index.products.add_document(doc)
    print(f"Product definition added for {product}")


def get_dem_tile_uris(bounding_box, product):
    # Validate the bounding_box
    if bounding_box is None:
        logging.warning(
            "No BBOX provided, running full extent... this will take a long time."
        )
        bounding_box = (-180, -90, 180, 90)
    else:
        bounding_box = bounding_box.split(",")
        if len(bounding_box) != 4:
            raise ValueError("bounding_box must be in the format: minx,miny,maxx,maxy")
        bounding_box = [float(x) for x in bounding_box]

    # Get the uris
    left = bounding_box[0]
    right = bounding_box[2]
    bottom = bounding_box[1]
    top = bounding_box[3]

    x_range = range(floor(left), ceil(right))
    y_range = range(floor(bottom), ceil(top))

    for x in x_range:
        for y in y_range:
            if x < 0:
                x_str = f"W{abs(x):03d}"
            else:
                x_str = f"E{x:03d}"
            if y < 0:
                y_str = f"S{abs(y):02d}"
            else:
                y_str = f"N{y:02d}"
            yield (
                URI_TEMPLATES[product].format(ns=y_str, ew=x_str),
                f"{x_str}_{y_str}",
            )


def process_uri_tile(
    uri_tile: Tuple[str, str, str],
    product: str,
    dc: Datacube,
    doc2ds: Doc2Dataset,
    update_if_exists: bool = True,
    archive_less_mature: bool = False,
    publish_action: str = None,
) -> Tuple[pystac.Item, str]:
    product_name = f"dem_{product}"
    uri, tile = uri_tile
    properties = {
        "odc:product": product_name,
        "odc:region_code": tile,
        "start_datetime": "1900-01-01",
        "end_datetime": "2100-01-01",
    }

    with rasterio.Env(aws_unsigned=True, GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR"):
        item = create_stac_item(
            uri,
            collection=product_name,
            with_proj=True,
            properties=properties,
            asset_media_type=pystac.MediaType.COG,
            asset_name="elevation",
        )

    index_update_dataset(
        stac_transform(item.to_dict()),
        uri,
        dc,
        doc2ds,
        update_if_exists=update_if_exists,
        allow_unsafe=True,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        stac_doc=item.to_dict(),
    )

    return True


def cop_dem_to_dc(
    dc: Datacube,
    product: str,
    bounding_box,
    limit: int,
    update: bool,
    n_workers: int = 100,
    archive_less_mature=None,
    publish_action=None,
) -> Tuple[int, int]:
    doc2ds = Doc2Dataset(dc.index)

    # Get a generator of (uris)
    uris_tiles = list(get_dem_tile_uris(bounding_box, product))
    if limit:
        uris_tiles = uris_tiles[0:limit]

    # Do the indexing of all the things
    success = 0
    failure = 0
    skipped = 0

    sys.stdout.write(f"Starting Cop DEM indexing with {n_workers} workers...\n")

    with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
        future_to_uri = {
            executor.submit(
                process_uri_tile,
                uri_tile,
                product,
                dc,
                doc2ds,
                update_if_exists=update,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
            ): uri_tile[0]
            for uri_tile in uris_tiles
        }
        for future in concurrent.futures.as_completed(future_to_uri):
            uri = future_to_uri[future]
            try:
                _ = future.result()
                success += 1
                if success % 10 == 0:
                    sys.stdout.write(f"\rAdded {success} datasets...")
            except SkippedException:
                skipped += 1
            except rasterio.errors.RasterioIOError:
                logging.info("Couldn't read file %s", uri, exc_info=True)
            except Exception:  # pylint:disable=broad-except
                logging.exception("Failed to handle uri %s", uri)
                failure += 1
    sys.stdout.write("\r")

    return success, failure, skipped


@click.command("cop-dem-to-dc")
@limit
@update_if_exists_flag
@bbox
@statsd_setting
@archive_less_mature
@publish_action
@click.option(
    "--product",
    default="cop_30",
    help="Product to add to the index, either cop_30 or cop_90",
)
@click.option(
    "--add-product",
    is_flag=True,
    default=False,
    help="If set, add the product too",
)
@click.option(
    "--workers",
    default=20,
    type=int,
    help="Number of threads to use to process, default 20",
)
def cli(
    limit,
    update_if_exists,
    bbox,
    statsd_setting,
    product,
    add_product,
    workers,
    archive_less_mature,
    publish_action,
):
    """
    Index the Copernicus DEM automatically.
    """
    if product not in PRODUCTS:
        raise ValueError(
            f"Unknown product {product}, must be one of {' '.join(PRODUCTS)}"
        )

    dc = Datacube()

    if add_product:
        add_cop_dem_product(dc, product)

    print(f"Indexing Copernicus DEM for {product} with bounding box of {bbox}")

    added, failed, skipped = cop_dem_to_dc(
        dc,
        product,
        bbox,
        limit,
        update_if_exists,
        n_workers=workers,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
    )

    print(
        f"Added {added} Datasets, failed {failed} Datasets, skipped {skipped} Datasets"
    )

    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:cop_dem_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:cop_dem_to_dc", "action:failed"], statsd_setting
        )
        statsd_gauge_reporting(
            skipped, ["app:cop_dem_to_dc", "action:skipped"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()



""" Index datasets from tar arachive
"""

import sys

import click
import datacube
from datacube.utils.changes import allow_any
from odc.io.tar import tar_doc_stream, tar_mode
from odc.io.timer import RateEstimator

from ._docs import from_yaml_doc_stream
from ._stac import stac_transform


def from_tar_file(tarfname, index, mk_uri, mode, doc_transform=None, **kwargs):
    """returns a sequence of tuples where each tuple is either

    (ds, None) or (None, error_message)
    """

    def untar(tarfname, mk_uri):
        for doc_name, doc in tar_doc_stream(tarfname, mode=mode):
            yield mk_uri(doc_name), doc

    return from_yaml_doc_stream(
        untar(tarfname, mk_uri), index, transform=doc_transform, **kwargs
    )


@click.command("index_from_tar")
@click.option("--env", "-E", type=str, help="Datacube environment name")
@click.option(
    "--product",
    "-p",
    "product_names",
    help=(
        "Only match against products specified with this option, "
        "you can supply several by repeating this option with a new product name"
    ),
    multiple=True,
)
@click.option(
    "--exclude-product",
    "-x",
    "exclude_product_names",
    help=(
        "Attempt to match to all products in the DB except for products "
        "specified with this option, "
        "you can supply several by repeating this option with a new product name"
    ),
    multiple=True,
)
@click.option(
    "--auto-add-lineage/--no-auto-add-lineage",
    is_flag=True,
    default=True,
    help=(
        "Default behaviour is to automatically add lineage datasets if they are missing from the database, "
        "but this can be disabled if lineage is expected to be present in the DB, "
        "in this case add will abort when encountering missing lineage dataset"
    ),
)
@click.option(
    "--verify-lineage/--no-verify-lineage",
    is_flag=True,
    default=True,
    help=(
        "Lineage referenced in the metadata document should be the same as in DB, "
        "default behaviour is to skip those top-level datasets that have lineage data "
        "different from the version in the DB. This option allows omitting verification step."
    ),
)
@click.option(
    "--ignore-lineage",
    help="Pretend that there is no lineage data in the datasets being indexed",
    is_flag=True,
    default=False,
)
@click.option(
    "--update", help="Update datasets rather than add", is_flag=True, default=False
)
@click.option(
    "--gzip",
    is_flag=True,
    help="Input is compressed with gzip (needed when reading from stdin)",
)
@click.option(
    "--xz",
    is_flag=True,
    help="Input is compressed with xz (needed when reading from stdin)",
)
@click.option(
    "--protocol",
    type=str,
    default="s3",
    show_default=True,
    help="Override the protocol for working with data in other environments, i.e gs",
)
@click.option(
    "--stac",
    is_flag=True,
    default=False,
    show_default=True,
    help="Expect STAC 1.0 metadata and attempt to transform to ODC EO3 metadata",
)
@click.argument("input_fname", type=str, nargs=-1)
def cli(
    input_fname,
    env,
    product_names,
    exclude_product_names,
    auto_add_lineage,
    verify_lineage,
    ignore_lineage,
    update,
    gzip,
    xz,
    protocol,
    stac,
):

    # Ensure :// is present in prefix
    prefix = protocol.rstrip("://") + "://"
    if prefix.startswith("file"):
        prefix = prefix + "/"

    if ignore_lineage:
        auto_add_lineage = False

    ds_resolve_args = dict(
        products=product_names,
        exclude_products=exclude_product_names,
        fail_on_missing_lineage=not auto_add_lineage,
        verify_lineage=verify_lineage,
        skip_lineage=ignore_lineage,
    )

    allowed_changes = {(): allow_any}

    transform = None
    if stac:
        transform = stac_transform

    def mk_uri(name):
        return prefix + name

    def report_error(msg):
        print(msg, file=sys.stderr)

    def process_file(filename, index, fps, mode=None, n_failed=0, doc_transform=None):
        for ds, err in from_tar_file(
            filename,
            index,
            mk_uri,
            mode=mode,
            doc_transform=doc_transform,
            **ds_resolve_args,
        ):
            if ds is not None:
                try:
                    if update:
                        index.datasets.update(ds, allowed_changes)
                    else:
                        index.datasets.add(ds, with_lineage=auto_add_lineage)

                except Exception as e:
                    n_failed += 1
                    report_error(str(e))
            else:
                n_failed += 1
                report_error(err)

            fps()

            if fps.every(10):
                print(".", end="", flush=True)

            if fps.every(100):
                print(" {} F:{:d}".format(str(fps), n_failed))

        return n_failed

    dc = datacube.Datacube(env=env)

    if len(input_fname) == 0:
        input_fname = ("-",)

    n_failed = 0
    fps = RateEstimator()
    mode = None

    for filename in input_fname:
        print(f"indexing {filename}")
        if filename == "-":
            if sys.stdin.isatty():
                report_error("Requesting to read from stdin but not redirecting input?")
                sys.exit(1)
            filename = sys.stdin.buffer
            mode = tar_mode(is_pipe=True, gzip=gzip, xz=xz)

        n_failed = process_file(
            filename,
            dc.index,
            fps,
            mode=mode,
            n_failed=n_failed,
            doc_transform=transform,
        )

    if n_failed > 0:
        report_error("**WARNING** there were failures: {}".format(n_failed))


if __name__ == "__main__":
    cli()



"""
Metadata Transformation from old format to new format:
Example usage:
 python export_md.py --datacube-config $HOME/.datacube.conf transform --product aster_l1t_swir
    --config metadata_transform_config.yaml --output-dir /g/data/u46/users/aj9439/metadata
    --limit 10
The config file contains the grid mappings to band names:
    grids:
      default: ['3', '4', '5']
      swir: ['1', '2']
"""
import logging
from pathlib import Path

import click
import yaml
from datacube import Datacube
from datacube.storage import BandInfo
from datacube.testutils.io import native_geobox

_LOG = logging.getLogger(__name__)


@click.group(help=__doc__)
@click.option(
    "--datacube-config",
    "-c",
    help="Pass the configuration file to access the database",
    type=click.Path(exists=True),
)
@click.pass_context
def cli(ctx, datacube_config):
    """Specify datacube index to be used for the given datacube config"""
    ctx.obj = Datacube(config=datacube_config).index


@cli.command()
@click.option("--product", required=True, help="Which product?")
@click.option(
    "--config", required=True, help="The configuration file for transformation"
)
@click.option(
    "--output-dir", required=True, help="New metadata yaml file is written to this dir"
)
@click.option("--limit", help="maximum number of datasets to process")
@click.pass_obj
def transform(index, product, config, output_dir, limit):

    # Get the product
    dataset_type = index.products.get_by_name(product)

    with open(config, "r") as config_file:
        cfg = yaml.load(config_file) or dict()

    # Is this a ingested product?
    if dataset_type.grid_spec is not None:
        transform_ingested_datasets(index, product, cfg, Path(output_dir), limit)
    else:
        transform_indexed_datasets(index, product, cfg, Path(output_dir), limit)


def transform_ingested_datasets(index, product, config, output_dir, limit):
    """
    Transform the metadata of ingested product. The product-wide fixed sections
    of metadata such as 'grids' is computed just once.
    """

    dataset_ids = index.datasets.search_returning(
        limit=limit, field_names=("id",), product=product
    )

    grids_done = False
    for dataset_id in dataset_ids:

        dataset = index.datasets.get(dataset_id.id, include_sources=True)

        if not dataset.uris:
            _LOG.warn("Empty uris or No uris (skippins): %s", dataset_id)
            continue

        if not grids_done:
            # setup grids
            grids = get_grids(dataset, config.get("grids"))

            # got to try to compute grids until shape is not default [1, 1]
            grids_done = all(
                [[1, 1] != grid["shape"] for _, grid in grids["grids"].items()]
            )

        dataset_sections = (grids,) + _variable_sections_of_metadata(dataset, config)
        _make_and_write_dataset(get_output_file(dataset, output_dir), *dataset_sections)


def transform_indexed_datasets(index, product, config, output_dir, limit):
    """
    Transform metadata of an indexed product. All sections of metadata are computed
    per dataset.
    """

    for dataset_id in index.datasets.search_returning(
        limit=limit, field_names=("id",), product=product
    ):

        dataset = index.datasets.get(dataset_id.id, include_sources=True)

        if not dataset.uris:
            _LOG.warn("Empty or No uris (skipping): %s", dataset_id)
            continue

        grids = get_grids(dataset, config.get("grids"))

        dataset_sections = (grids,) + _variable_sections_of_metadata(dataset, config)
        _make_and_write_dataset(get_output_file(dataset, output_dir), *dataset_sections)


def get_output_file(dataset, output_dir):
    """
    Compute the output metadata file name for the given dataset.
    """

    out_file_name = str(dataset.id) + "-metadata.yaml"

    return output_dir / out_file_name


def _variable_sections_of_metadata(dataset, config):
    """
    Compute variable sections (i.e. those sections that vary per dataset)
    """
    new_dataset = {
        "id": str(dataset.id),
        "crs": "EPSG:" + str(dataset.crs.epsg),
        "location": [uri for uri in dataset.uris],
        "file_format": dataset.metadata_doc.get("format", ""),
    }

    return (
        new_dataset,
        get_geometry(dataset),
        get_measurements(dataset, config.get("grids")),
        get_properties(dataset),
        get_lineage(dataset),
    )


def _make_and_write_dataset(out_file_name, *args):
    """
    Assemble the metadata sections and write out.
    """

    dataset = dict()
    for arg in args:
        dataset.update(arg)

    with open(out_file_name, "w") as out_file:
        yaml.dump(dataset, out_file, default_flow_style=False)


def get_geometry(dataset):
    """
    Extract and return geometry coordinates as a list in a dictionary:
    returns
    {
      'geometry': {
                    'type' : 'Polygon',
                    'coordinates': [...]
                  }
    }
    or
    empty dictionary
    """

    valid_data = dataset._gs.get("valid_data")

    return {"geometry": valid_data} if valid_data else dict()


def get_grids(dataset, band_grids=None):
    """
    'band_grids' specify bands for each grid as in:
    {
        default: [swir1, swir2]
        ir: [ir1, ir2]
    }
    All the bands not included in the 'band_grids' considered 'default'
    Returns grids as in
    {
     grids:
        default:
            shape: [7731, 7621]
            transform: [30.0, 0.0, 306285.0, 0.0, -30.0, -1802085.0, 0, 0, 1]
        ir:
            shape: [3865, 3810]
            transform: [60.0, 0.0, 306285.0, 0.0, -60.0, -1802085.0, 0, 0, 1]
    }
    """

    if not band_grids:
        # Assume all measurements belong to default grid
        shape, trans = get_shape_and_transform(dataset, dataset.measurements)

        return {"grids": {"default": {"shape": list(shape), "transform": list(trans)}}}
    else:
        grids = dict()
        for grid_name in band_grids:

            shape, trans = get_shape_and_transform(dataset, band_grids[grid_name])

            grids[grid_name] = {"shape": list(shape), "transform": list(trans)}

        if not band_grids.get("default"):

            specified_bands = set()
            for grid in band_grids:
                specified_bands.update(band_grids[grid])
            all_bands = set(list(dataset.measurements))

            default_bands = all_bands - specified_bands

            if bool(default_bands):
                shape, trans = get_shape_and_transform(dataset, default_bands)

                grids["default"] = {"shape": list(shape), "transform": list(trans)}
        return {"grids": grids}


def get_shape_and_transform(dataset, measurements):
    """
    Get shape and transform for the given set of measurements. It try
    each measurement until it succeeds.
    """

    for m in measurements:
        try:
            geo = native_geobox(dataset, [m])
        except Exception:
            _LOG.warn("Failed to compute shape and transform %s", m)
            continue
        return geo.shape, geo.transform

    # All the bands failed use default shape [1,1]
    _LOG.warn("All measurements failed to compute shape and transform %s", measurements)
    return [1, 1], dataset.transform


def get_measurements(dataset, band_grids=None):
    """
    Extract and return measurement paths in a dictionary:
    Returns
    {
      'measurements':
      {
        'coastal_aerosol': {
                             'path': path_name1,
                             'band': band_number,
                             'layer': null or 'layer_name'
                             'grid': 'ir'
                           },
        ...
      }
    }
    """
    grids_map = (
        {m: grid for grid in band_grids for m in band_grids[grid] if grid != "default"}
        if band_grids
        else dict()
    )

    measurements = dict()
    for m in dataset.measurements:

        if m not in dataset.type.measurements:
            _LOG.warn("Measurement not in product definition (skipping): %s", m)
            continue

        band_info = BandInfo(dataset, m)
        measurements[m] = {"path": dataset.measurements[m]["path"]}

        if band_info.band and band_info.band != 1:
            measurements[m]["band"] = band_info.band

        if band_info.layer is not None:
            measurements[m]["layer"] = band_info.layer

        if grids_map.get(m):
            measurements[m]["grid"] = grids_map[m]

    return {"measurements": measurements}


def get_properties(dataset, property_offsets=None):
    """
    Extract properties and return values in a dictionary:
    {
      'properties':
      {
        'datetime': time,
        'odc:creation_datetime': creation_time,
        ...
      }
    }
    """
    props = dict()
    props["datetime"] = dataset.center_time
    props["odc:processing_datetime"] = dataset.indexed_time

    return {"properties": props}


def get_lineage(dataset):
    """
    Extract immediate parents.
    {
      'lineage':
      {
        'nbar': [id1, id2],
        'pq': [id3]
        ...
      }
    }
    """
    lineage = dict()
    for classifier in dataset.sources:
        lineage[classifier] = str(dataset.sources[classifier].id)
    return {"lineage": lineage}


if __name__ == "__main__":
    cli()



import json
import logging
from pathlib import Path
from typing import Generator, Optional

import click
import datacube
import yaml
from datacube.index.hl import Doc2Dataset
from odc.apps.dc_tools._stac import stac_transform
from odc.apps.dc_tools.utils import (
    allow_unsafe,
    archive_less_mature,
    index_update_dataset,
    update_if_exists_flag,
    publish_action,
    statsd_setting,
    statsd_gauge_reporting,
    transform_stac,
)
from odc.apps.dc_tools._stac import stac_transform
import logging


import yaml

logging.basicConfig(
    level=logging.WARNING,
    format="%(asctime)s: %(levelname)s: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S",
)


@click.command("fs-to-dc")
@click.argument("input_directory", type=str, nargs=1)
@update_if_exists_flag
@allow_unsafe
@archive_less_mature
@transform_stac
@statsd_setting
@publish_action
@click.option(
    "--glob",
    default=None,
    help="File system glob to use, defaults to **/*.yaml or **/*.json for STAC.",
)
def cli(
    input_directory,
    update_if_exists,
    allow_unsafe,
    stac,
    statsd_setting,
    glob,
    archive_less_mature,
    publish_action,
):

    dc = datacube.Datacube()
    doc2ds = Doc2Dataset(dc.index)

    if glob is None:
        glob = "**/*.json" if stac else "**/*.yaml"

    files_to_process = Path(input_directory).glob(glob)

    added, failed = 0, 0

    for in_file in files_to_process:
        with in_file.open() as f:
            try:
                if in_file.suffix in (".yml", ".yaml"):
                    metadata = yaml.safe_load(f)
                else:
                    metadata = json.load(f)
                # Do the STAC Transform if it's flagged
                stac_doc = None
                if stac:
                    stac_doc = metadata
                    metadata = stac_transform(metadata)
                index_update_dataset(
                    metadata,
                    in_file.absolute().as_uri(),
                    dc=dc,
                    doc2ds=doc2ds,
                    update_if_exists=update_if_exists,
                    allow_unsafe=allow_unsafe,
                    archive_less_mature=archive_less_mature,
                    publish_action=publish_action,
                    stac_doc=stac_doc,
                )
                added += 1
            except Exception:  # pylint: disable=broad-except
                logging.exception("Failed to add dataset %s", in_file)
                failed += 1

    logging.info("Added %s and failed %s datasets.", added, failed)
    if statsd_setting:
        statsd_gauge_reporting(added, ["app:fs_to_dc", "action:added"], statsd_setting)
        statsd_gauge_reporting(
            failed, ["app:fs_to_dc", "action:failed"], statsd_setting
        )


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""Index datasets found from an SQS queue into Postgres
"""
import json
import logging
import sys
import uuid
from pathlib import PurePath
from typing import Tuple

import boto3
import click
import pandas as pd
import requests
from botocore import UNSIGNED
from botocore.config import Config
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.utils import documents
from odc.apps.dc_tools.utils import (
    IndexingException,
    SkippedException,
    allow_unsafe,
    archive,
    fail_on_missing_lineage,
    index_update_dataset,
    limit,
    no_sign_request,
    skip_lineage,
    statsd_setting,
    statsd_gauge_reporting,
    transform_stac,
    transform_stac_absolute,
    archive_less_mature,
    update_flag,
    update_if_exists_flag,
    verify_lineage,
    publish_action,
)
from odc.aws.queue import get_messages, publish_to_topic
from toolz import dicttoolz
from yaml import safe_load

from ._stac import stac_transform, stac_transform_absolute, ds_to_stac

# Added log handler
logging.basicConfig(level=logging.WARNING, handlers=[logging.StreamHandler()])


def extract_metadata_from_message(message):
    try:
        body = json.loads(message.body)
        metadata = json.loads(body["Message"])
    except (KeyError, json.JSONDecodeError) as e:
        raise IndexingException(
            f"Failed to load metadata from the SQS message due to error: {e}"
        )

    if metadata:
        return metadata
    else:
        raise IndexingException("Failed to load metadata from the SQS message")


def handle_json_message(metadata, transform, odc_metadata_link):
    odc_yaml_uri = None
    uri = None

    if odc_metadata_link:
        if odc_metadata_link.startswith("STAC-LINKS-REL:"):
            rel_val = odc_metadata_link.replace("STAC-LINKS-REL:", "")
            odc_yaml_uri = get_uri(metadata, rel_val)
        else:
            # if odc_metadata_link is provided, it will look for value with dict path provided
            odc_yaml_uri = dicttoolz.get_in(odc_metadata_link.split("/"), metadata)

        # if odc_yaml_uri exist, it will load the metadata content from that URL
        if odc_yaml_uri:
            try:
                content = requests.get(odc_yaml_uri).content
                metadata = documents.parse_yaml(content)
                uri = odc_yaml_uri
            except requests.RequestException as err:
                raise IndexingException(
                    f"Failed to load metadata from the link provided -  {err}"
                )
        else:
            raise IndexingException("ODC EO3 metadata link not found")
    else:
        # if no odc_metadata_link provided, it will look for metadata dict "href" value with "rel==self"
        uri = get_uri(metadata, "self")

    if transform:
        metadata = transform(metadata)

    return metadata, uri


def handle_bucket_notification_message(
    message, metadata: dict, record_path: tuple, no_sign_request: bool = False
) -> Tuple[dict, str]:
    """[summary]

    Args:
        message (Message resource)
        metadata (dict): [description]
        record_path (tuple): [PATH for selectingthe s3 key path from the JSON message document]

    Raises:
        IndexingException: [Catch s3 ]

    Returns:
        Tuple[dict, str]: [description]
    """
    data = None
    uri = None

    if metadata.get("Records"):
        for record in metadata.get("Records"):
            bucket_name = dicttoolz.get_in(["s3", "bucket", "name"], record)
            key = dicttoolz.get_in(["s3", "object", "key"], record)

            # Check for bucket name and key, and fail if there isn't one
            if not (bucket_name and key):
                # Not deleting this message, as it's non-conforming. Check this logic
                raise IndexingException(
                    "No bucket name or key in message, are you sure this is a bucket notification?"
                )

            # If you specific a list of record paths, and there's no
            # match in them for the key, then we skip this one forever
            if record_path is not None and not any(
                PurePath(key).match(p) for p in record_path
            ):
                logging.warning(
                    "Key: %s not in specified list of record_paths, deleting message from the queue.",
                    key,
                )
                # This will return Nones, which will flag the message to be ignored
                return None, None

            # We have enough information to proceed, get the key and extract
            # the contents...
            try:
                uri = f"s3://{bucket_name}/{key}"
                if no_sign_request:
                    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
                    data = s3.get_object(Bucket=bucket_name, Key=key)
                    contents = data["Body"].read()
                    data = safe_load(contents.decode("utf-8"))
                else:
                    s3 = boto3.resource("s3")
                    obj = s3.Object(bucket_name, key).get(
                        ResponseCacheControl="no-cache"
                    )
                    data = safe_load(obj["Body"].read())
            except Exception as e:
                raise IndexingException(
                    "Exception thrown when trying to load s3 object"
                ) from e
    else:
        raise IndexingException(
            "Attempted to get metadata from record when no record key exists in message."
        )

    return data, uri


def get_uri(metadata, rel_value):
    uri = None
    for link in metadata.get("links"):
        rel = link.get("rel")
        if rel and rel == rel_value:
            uri = link.get("href")
    return uri


def do_archiving(metadata, dc: Datacube, publish_action):
    dataset_id = uuid.UUID(metadata.get("id"))
    if dataset_id:
        dc.index.datasets.archive([dataset_id])
        if publish_action:
            publish_to_topic(
                arn=publish_action,
                action="ARCHIVED",
                stac=ds_to_stac(dc.index.datasets.get(dataset_id)),
            )
    else:
        raise IndexingException("Failed to get an ID from the message, can't archive.")


def queue_to_odc(
    queue,
    dc: Datacube,
    products: list,
    record_path=None,
    transform=None,
    limit=None,
    update=False,
    update_if_exists=False,
    no_sign_request=False,
    archive=False,
    allow_unsafe=False,
    odc_metadata_link=False,
    region_code_list_uri=None,
    archive_less_mature=None,
    publish_action=None,
    **kwargs,
) -> Tuple[int, int, int]:

    ds_success = 0
    ds_failed = 0
    ds_skipped = 0

    region_codes = None
    if region_code_list_uri:
        try:
            region_codes = set(
                pd.read_csv(region_code_list_uri, header=None).values.ravel()
            )
        except FileNotFoundError:
            logging.exception("Could not find region_code file")
        if len(region_codes) == 0:
            raise IndexingException(
                f"Region code list is empty, please check the list at: {region_code_list_uri}"
            )

    doc2ds = Doc2Dataset(dc.index, products=products, **kwargs)

    # This is a generator of messages
    messages = get_messages(queue, limit)

    for message in messages:
        try:
            # Extract metadata from message
            metadata = extract_metadata_from_message(message)
            stac_doc = None
            if archive:
                # Archive metadata
                do_archiving(metadata, dc, publish_action)
            else:
                if not record_path:
                    if transform:
                        stac_doc = metadata
                    # Extract metadata and URI from a STAC or similar
                    # json structure for indexing
                    metadata, uri = handle_json_message(
                        metadata, transform, odc_metadata_link
                    )
                else:
                    # Extract metadata from an S3 bucket notification
                    # or similar for indexing
                    metadata, uri = handle_bucket_notification_message(
                        message, metadata, record_path, no_sign_request=no_sign_request
                    )

                # If we have a region_code filter, do it here
                if region_code_list_uri:
                    region_code = dicttoolz.get_in(
                        ["properties", "odc:region_code"], metadata
                    )
                    if region_code not in region_codes:
                        # We don't want to keep this one, so flag it so
                        # it's not indexed by is still deleted.
                        metadata = None
                        uri = None
                        logging.warning(
                            "Region code %s not in list of allowed region codes, ignoring this dataset.",
                            region_code,
                        )

                # Index the dataset
                if metadata is not None and uri is not None:
                    try:
                        index_update_dataset(
                            metadata,
                            uri,
                            dc,
                            doc2ds,
                            update=update,
                            update_if_exists=update_if_exists,
                            allow_unsafe=allow_unsafe,
                            archive_less_mature=archive_less_mature,
                            publish_action=publish_action,
                            stac_doc=stac_doc,
                        )
                        ds_success += 1
                    except SkippedException:
                        ds_skipped += 1
                else:
                    logging.warning("Found None for metadata and uri, skipping")
                    ds_skipped += 1

            # Success, so delete the message.
            message.delete()
        except IndexingException:
            logging.exception("Failed to handle SQS message")
            ds_failed += 1

    return ds_success, ds_failed, ds_skipped


@click.command("sqs-to-dc")
@skip_lineage
@fail_on_missing_lineage
@verify_lineage
@transform_stac
@transform_stac_absolute
@update_flag
@update_if_exists_flag
@allow_unsafe
@archive
@limit
@no_sign_request
@statsd_setting
@click.option(
    "--odc-metadata-link",
    default=None,
    help="Expect metadata doc with ODC EO3 metadata link. "
    "Either provide '/' separated path to find metadata link in a provided "
    "metadata doc e.g. 'foo/bar/link', or if metadata doc is STAC, "
    "provide 'rel' value of the 'links' object having "
    "metadata link. e.g. 'STAC-LINKS-REL:odc_yaml'",
)
@click.option(
    "--record-path",
    default=None,
    multiple=True,
    help="Filtering option for s3 path, i.e. 'L2/sentinel-2-nrt/S2MSIARD/*/*/ARD-METADATA.yaml'",
)
@click.option(
    "--region-code-list-uri",
    default=None,
    help="A path to a list (one item per line, in txt or gzip format) of valide region_codes to include",
)
@archive_less_mature
@publish_action
@click.argument("queue_name", type=str, nargs=1)
@click.argument("product", type=str, nargs=1)
def cli(
    skip_lineage,
    fail_on_missing_lineage,
    verify_lineage,
    stac,
    absolute,
    update,
    update_if_exists,
    allow_unsafe,
    archive,
    limit,
    statsd_setting,
    no_sign_request,
    odc_metadata_link,
    record_path,
    region_code_list_uri,
    archive_less_mature,
    publish_action,
    queue_name,
    product,
):
    """Iterate through messages on an SQS queue and add them to datacube"""

    transform = None
    if stac:
        if absolute:
            transform = stac_transform_absolute
        else:
            transform = stac_transform

    candidate_products = product.split()

    sqs = boto3.resource("sqs")
    queue = sqs.get_queue_by_name(QueueName=queue_name)

    # Do the thing
    dc = Datacube()
    success, failed, skipped = queue_to_odc(
        queue,
        dc,
        candidate_products,
        skip_lineage=skip_lineage,
        fail_on_missing_lineage=fail_on_missing_lineage,
        verify_lineage=verify_lineage,
        transform=transform,
        limit=limit,
        update=update,
        no_sign_request=no_sign_request,
        update_if_exists=update_if_exists,
        archive=archive,
        allow_unsafe=allow_unsafe,
        record_path=record_path,
        odc_metadata_link=odc_metadata_link,
        region_code_list_uri=region_code_list_uri,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
    )

    result_msg = ""
    if update:
        result_msg += f"Updated {success} Dataset(s), "
        if statsd_setting:
            statsd_gauge_reporting(
                success, ["app:sqs_to_dc", "action:update"], statsd_setting
            )
    elif archive:
        result_msg += f"Archived {success} Dataset(s), "
        if statsd_setting:
            statsd_gauge_reporting(
                success, ["app:sqs_to_dc", "action:archived"], statsd_setting
            )
    else:
        result_msg += f"Added {success} Dataset(s), "
        if statsd_setting:
            statsd_gauge_reporting(
                success, ["app:sqs_to_dc", "action:added"], statsd_setting
            )
    result_msg += f"Failed {failed} Dataset(s), Skipped {skipped} Dataset(s)"
    print(result_msg)

    if statsd_setting:
        statsd_gauge_reporting(
            failed, ["app:sqs_to_dc", "action:failed"], statsd_setting
        )
        statsd_gauge_reporting(
            skipped, ["app:sqs_to_dc", "action:skipped"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()



"""These should probably be in datacube library."""
import json
import sys
from typing import Sequence, Union
from uuid import UUID, uuid5

from datacube.index.hl import Doc2Dataset
from datacube.utils.documents import parse_yaml

# Some random UUID to be ODC namespace
ODC_NS = UUID("6f34c6f4-13d6-43c0-8e4e-42b6c13203af")


def odc_uuid(
    algorithm: str,
    algorithm_version: str,
    sources: Sequence[Union[UUID, str]],
    deployment_id: str = "",
    **other_tags,
) -> UUID:
    """
    Generate deterministic UUID for a derived Dataset.

    :param algorithm: Name of the algorithm
    :param algorithm_version: Version string of the algorithm
    :param sources: Sequence of input Dataset UUIDs
    :param deployment_id: Some sort of identifier for installation that performs
                          the run, for example Docker image hash, or dea module version on NCI.
    :param **other_tags: Any other identifiers necessary to uniquely identify dataset
    """
    tags = [f"{k}={str(v)}" for k, v in other_tags.items()]

    stringified_sources = (
        [str(algorithm), str(algorithm_version), str(deployment_id)]
        + sorted(tags)
        + [str(u) for u in sorted(sources)]
    )

    srcs_hashes = "\n".join(s.lower() for s in stringified_sources)
    return uuid5(ODC_NS, srcs_hashes)


def from_metadata_stream(metadata_stream, index, **kwargs):
    """
    Raw metadata to Dataset stream converter.

    Given a stream of (uri, metadata) tuples convert them into Datasets, using
    supplied index and options for Doc2Dataset.


    **kwargs**:
    - skip_lineage
    - verify_lineage
    - fail_on_missing_lineage
    - products
    - exclude_products

    returns a sequence of tuples where each tuple is either

        (Dataset, None) or (None, error_message)
    """
    doc2ds = Doc2Dataset(index, **kwargs)

    for uri, metadata in metadata_stream:
        if metadata is None:
            yield (None, f"Error: empty doc {uri}")
        else:
            ds, err = doc2ds(metadata, uri)
            if ds is not None:
                yield (ds, None)
            else:
                yield (None, f"Error: {uri}, {err}")


def parse_doc_stream(doc_stream, on_error=None, transform=None):
    """
    Replace doc bytes/strings with parsed dicts.

       Stream[(uri, bytes)] -> Stream[(uri, dict)]


    :param doc_stream: sequence of (uri, doc: bytes|string)
    :param on_error: Callback uri, doc -> None
    :param transform: dict -> dict if supplied also apply further transform on parsed document

    On output doc is replaced with python dict parsed from yaml, or with None
    if parsing/transform error occurred.
    """
    for uri, doc in doc_stream:
        try:
            if uri.endswith(".json"):
                metadata = json.loads(doc)
            else:
                metadata = parse_yaml(doc)

            if transform is not None:
                metadata = transform(metadata)
        except Exception:  # pylint: disable=broad-except
            if on_error is not None:
                on_error(uri, doc)
            metadata = None

        yield uri, metadata


def from_yaml_doc_stream(doc_stream, index, logger=None, transform=None, **kwargs):
    """
    Stream of yaml documents to a stream of Dataset results.

    Stream[(path, bytes|str)] -> Stream[(Dataset, None)|(None, error_message)]

    :param doc_stream: sequence of (uri, doc: byges|string)
    :param on_error: Callback uri, doc -> None
    :param logger:  Logger object for printing errors or None
    :param transform: dict -> dict if supplied also apply further transform on parsed document
    :param kwargs: passed on to from_metadata_stream

    """

    def on_parse_error(uri, doc):
        # pylint: disable=unused-argument
        if logger is not None:
            logger.error(f"Failed to parse: {uri}")
        else:
            print(f"Failed to parse: {uri}", file=sys.stderr)

    metadata_stream = parse_doc_stream(
        doc_stream, on_error=on_parse_error, transform=transform
    )
    return from_metadata_stream(metadata_stream, index, **kwargs)



"""
Tools for STAC to EO3 translation
"""
import math
from pathlib import Path
from typing import Any, Dict, Optional, Tuple
from uuid import UUID

import numpy
from datacube.utils.geometry import Geometry
from toolz import get_in
from urlpath import URL

from ._docs import odc_uuid

from eodatasets3.serialise import from_doc
from eodatasets3.stac import to_stac_item
from datacube.model import Dataset

Document = Dict[str, Any]

# This is an old hack, should be refactored out
DEA_LANDSAT_PRODUCTS = ["ga_ls8c_ard_3", "ga_ls7e_ard_3", "ga_ls8t_ard_3"]

# This is hack for not changing the current behavior of DEAfrica sentinel-2
# It is not ideal but to remain the impact minimum
TO_BE_HARD_CODED_COLLECTION = [
    "s2_l2a",
    "sentinel_s2_l2a_cogs",
    "sentinel-s2-l2a-cogs",
    "sentinel-2-l2a",
]

# Mapping between EO3 field names and STAC properties object field names
MAPPING_STAC_TO_EO3 = {
    "end_datetime": "dtr:end_datetime",
    "start_datetime": "dtr:start_datetime",
    "gsd": "eo:gsd",
    "instruments": "eo:instrument",
    "platform": "eo:platform",
    "constellation": "eo:constellation",
    "view:off_nadir": "eo:off_nadir",
    "view:azimuth": "eo:azimuth",
    "view:sun_azimuth": "eo:sun_azimuth",
    "view:sun_elevation": "eo:sun_elevation",
}

# Add more here as they are discovered
CANDIDATE_REGION_CODES = ["odc:region_code", "s2:mgrs_tile", "io:supercell_id"]


def _get_region_code(properties: Dict[str, Any]) -> str:
    # Check the default candidates
    for region_code_name in CANDIDATE_REGION_CODES:
        region_code = properties.get(region_code_name, None)
        if region_code is not None:
            return region_code

    # Landsat special case
    if region_code is None:
        row = get_in(["landsat:wrs_row"], properties, None)
        path = get_in(["landsat:wrs_path"], properties, None)
        try:
            if row is not None and path is not None:
                region_code = f"{int(path):03d}{int(row):03d}"
        except ValueError:
            pass

    return region_code


def _get_usgs_product_name(properties: Dict[str, Any]) -> str:
    platform = get_in(["platform"], properties)

    if platform == "LANDSAT_9":
        return "ls9-c2l2-sr"
    if platform == "LANDSAT_8":
        return "ls8-c2l2-sr"
    elif platform == "LANDSAT_7":
        return "ls7-c2l2-sr"
    elif platform == "LANDSAT_5":
        return "ls5-c2l2-sr"
    else:
        return None


def _stac_product_lookup(
    item: Document,
) -> Tuple[str, Optional[str], str, Optional[str], str]:
    properties = item["properties"]

    dataset_id: str = item["id"]
    dataset_label = item.get("title")

    # Get the simplest version of a product name
    product_name = properties.get("odc:product", None)

    # Try to get a known region_code
    region_code = _get_region_code(properties)

    default_grid = None

    # Maybe this should be the default product_name
    constellation = properties.get("constellation") or properties.get(
        "eo:constellation"
    )
    if constellation is not None:
        constellation = constellation.lower().replace(" ", "-")

    collection = item.get("collection")
    # Special case for USGS Landsat Collection 2
    if collection is not None and collection == "landsat-c2l2-sr":
        product_name = _get_usgs_product_name(properties)

    # It is an ugly hack without interrupting DEAfric sentinel 2
    if constellation is not None and product_name is None:
        if constellation == "sentinel-2" and collection in TO_BE_HARD_CODED_COLLECTION:
            dataset_id = properties.get("sentinel:product_id") or properties.get(
                "s2:granule_id", dataset_id
            )
            product_name = "s2_l2a"
            if region_code is None:
                # Let's try two options, and throw an exception if we still don't get it
                try:
                    # The 'mgrs' prefix (and STAC extension) started with STAC v1.0.0
                    region_code = (
                        f"{str(properties['proj:epsg'])[-2:]}"
                        f"{properties['mgrs:latitude_band']}"
                        f"{properties['mgrs:grid_square']}"
                    )
                except KeyError:
                    region_code = (
                        f"{str(properties['proj:epsg'])[-2:]}"
                        f"{properties['sentinel:latitude_band']}"
                        f"{properties['sentinel:grid_square']}"
                    )

            default_grid = "g10m"

    # If we still don't have a product name, use collection
    if product_name is None:
        product_name = collection
        if product_name is None:
            raise ValueError("Can't find product name from odc:product or collection.")

    # Product names can't have dashes in them
    product_name = product_name.replace("-", "_")

    if product_name in DEA_LANDSAT_PRODUCTS:
        self_href = _find_self_href(item)
        dataset_label = Path(self_href).stem.replace(".stac-item", "")
        default_grid = "g30m"

    # If the ID is not cold and numerical, assume it can serve a label.
    if (
        not dataset_label
        and not _check_valid_uuid(dataset_id)
        and not dataset_id.isnumeric()
    ):
        dataset_label = dataset_id

    return dataset_id, dataset_label, product_name, region_code, default_grid


def _find_self_href(item: Document) -> str:
    """
    Extracting product label from filename of the STAC document 'self' URL
    """
    self_uri = [
        link.get("href", "")
        for link in item.get("links", [])
        if link.get("rel") == "self"
    ]

    if len(self_uri) < 1:
        return None
    if len(self_uri) > 1:
        print("More than one links to 'self', returning the first one")
    return self_uri[0]


def _get_stac_bands(
    item: Document,
    default_grid: str,
    relative: bool = False,
    proj_shape: Optional[str] = None,
    proj_transform: Optional[str] = None,
) -> Tuple[Document, Document, Document]:
    # pylint: disable=too-many-locals
    bands = {}
    grids = {}
    accessories = {}

    self_link = _find_self_href(item)

    assets = item.get("assets", {})

    def _get_path(asset, force_relative=False):
        path = URL(asset["href"])
        if relative:
            try:
                if self_link is None:
                    raise ValueError
                path = path.relative_to(URL(self_link).parent)
            # Value error is raised if the path is not relative to the parent
            # or if the self link cannot be found.
            except ValueError:
                # If the path is not relative to the parent force_relative
                # is still used for data assets, due to a historical assumption.
                # TODO: Implement rewrite_assets (like in stac_to_dc) in all
                # tools so that this is no longer necessary.
                if force_relative:
                    path = path.name
                else:
                    pass

        return str(path)

    for asset_name, asset in assets.items():
        # If something's not a geotiff, make it an accessory
        # include thumbnails in accessories
        if "geotiff" not in asset.get("type", []) or "thumbnail" in asset.get(
            "roles", []
        ):
            accessories[asset_name] = {"path": _get_path(asset)}
            continue

        # If transform specified here in the asset it should override
        # the properties-specified transform.
        transform = asset.get("proj:transform") or proj_transform
        grid = f"g{transform[0]:g}m"

        # As per transform, shape here overrides properties
        shape = asset.get("proj:shape") or proj_shape

        if grid not in grids:
            grids[grid] = {
                "shape": shape,
                "transform": transform,
            }

        path = _get_path(asset, force_relative=True)
        band_index = asset.get("band", None)

        band_info = {"path": path}
        if band_index is not None:
            band_info["band"] = band_index

        # If we don't specify a default grid, label the first grid 'default'
        if not default_grid:
            default_grid = list(grids.keys())[0]

        if grid != default_grid:
            band_info["grid"] = grid

        bands[asset_name] = band_info

    if default_grid in grids:
        grids["default"] = grids.pop(default_grid)

    return bands, grids, accessories


def _geographic_to_projected(geometry, crs, precision=10):
    """Transform from WGS84 to the target projection, assuming Lon, Lat order"""
    geom = geometry.to_crs(crs, resolution=math.inf)

    def round_coords(c1, c2):
        return [round(coord, precision) for coord in [c1, c2]]

    if geom.is_valid:
        return geom.transform(round_coords)
    else:
        return None


def stac_transform_absolute(input_stac):
    return stac_transform(input_stac, relative=False)


def _convert_value_to_eo3_type(key: str, value):
    """
    Convert return type as per EO3 specification.
    Return type is String for "instrument" field in EO3 metadata.

    """
    if key == "instruments":
        if len(value) > 0:
            return "_".join([i.upper() for i in value])
        else:
            return None
    else:
        return value


def _get_stac_properties_lineage(input_stac: Document) -> Tuple[Document, Any]:
    """
    Extract properties and lineage field
    """
    properties = input_stac["properties"]
    prop = {
        MAPPING_STAC_TO_EO3.get(key, key): _convert_value_to_eo3_type(key, val)
        for key, val in properties.items()
    }

    creation_time = (
        properties.get("odc:processing_datetime")
        or properties.get("created")
        or properties.get("datetime")
    )
    if prop.get("odc:processing_datetime") is None and creation_time:
        prop["odc:processing_datetime"] = creation_time

    if prop.get("odc:file_format") is None:
        prop["odc:file_format"] = "GeoTIFF"

    # Extract lineage
    lineage = prop.pop("odc:lineage", None)

    return prop, lineage


def _check_valid_uuid(uuid_string: str) -> bool:
    """
    Check if provided uuid string is a valid UUID.
    """
    try:
        UUID(str(uuid_string))
        return True
    except ValueError:
        return False


def stac_transform(input_stac: Document, relative: bool = True) -> Document:
    """Takes in a raw STAC 1.0 dictionary and returns an ODC dictionary"""
    # pylint: disable=too-many-locals

    (
        dataset_id,
        dataset_label,
        product_name,
        region_code,
        default_grid,
    ) = _stac_product_lookup(input_stac)

    # Generating UUID for products not having UUID.
    # Checking if provided id is valid UUID.
    # If not valid, creating new deterministic uuid using odc_uuid function
    # based on product_name and product_label.
    # TODO: Verify if this approach to create UUID is valid.
    if _check_valid_uuid(input_stac["id"]):
        deterministic_uuid = input_stac["id"]
    else:
        if product_name in ["s2_l2a"]:
            deterministic_uuid = str(
                odc_uuid("sentinel-2_stac_process", "1.0.0", [dataset_id])
            )
        else:
            deterministic_uuid = str(
                odc_uuid(f"{product_name}_stac_process", "1.0.0", [dataset_id])
            )

    # Check for projection extension properties that are not in the asset fields.
    # Specifically, proj:shape and proj:transform, as these are otherwise
    # fetched in _get_stac_bands.
    properties = input_stac["properties"]
    proj_shape = properties.get("proj:shape")
    proj_transform = properties.get("proj:transform")
    # TODO: handle old STAC that doesn't have grid information here...
    bands, grids, accessories = _get_stac_bands(
        input_stac,
        default_grid,
        relative=relative,
        proj_shape=proj_shape,
        proj_transform=proj_transform,
    )

    stac_properties, lineage = _get_stac_properties_lineage(input_stac)

    epsg = properties["proj:epsg"]
    native_crs = f"epsg:{epsg}"

    # Transform geometry to the native CRS at an appropriate precision
    geometry = Geometry(input_stac["geometry"], "epsg:4326")
    if native_crs != "epsg:4326":
        # Arbitrary precisions, but should be fine
        pixel_size = get_in(["default", "transform", 0], grids, no_default=True)
        precision = 0
        if pixel_size < 0:
            precision = 6

        geometry = _geographic_to_projected(geometry, native_crs, precision)

    stac_odc = {
        "$schema": "https://schemas.opendatacube.org/dataset",
        "id": deterministic_uuid,
        "crs": native_crs,
        "grids": grids,
        "product": {"name": product_name.lower()},
        "properties": stac_properties,
        "measurements": bands,
        "lineage": {},
        "accessories": accessories,
    }
    if dataset_label:
        stac_odc["label"] = dataset_label

    if region_code:
        stac_odc["properties"]["odc:region_code"] = region_code

    if geometry:
        stac_odc["geometry"] = transform_geom_json_coordinates_to_list(geometry.json)

    if lineage:
        stac_odc["lineage"] = lineage

    return stac_odc


# TODO: This is a temporary fix
def transform_geom_json_coordinates_to_list(geom_json):
    geom_json["coordinates"] = numpy.array(geom_json["coordinates"]).tolist()
    return geom_json


def ds_to_stac(ds: Dataset) -> dict:
    """Get STAC document from dataset with eo3 metadata"""
    if ds.is_eo3:
        if not ds.uris:
            raise ValueError("Can't find dataset location")
        location = ds.uris[0]
        stac = to_stac_item(
            from_doc(ds.metadata_doc, skip_validation=True),
            location,
        )
        return stac
    else:
        raise ValueError(
            f"Cannot convert to STAC for dataset with metadata of type {ds.metadata_type}"
        )



"""Crawl Thredds for prefixes and fetch YAML's for indexing
and dump them into a Datacube instance
"""
import json
import logging
from typing import List, Optional

import click
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from odc.apps.dc_tools._stac import stac_transform
from odc.apps.dc_tools.utils import (
    allow_unsafe,
    archive_less_mature,
    index_update_dataset,
    statsd_gauge_reporting,
    statsd_setting,
    transform_stac,
    update_flag,
    update_if_exists_flag,
    publish_action,
)
from odc.azure import download_blob, find_blobs


def stream_blob_urls(account_url, container_name, credential, blobs: List[str]):
    for blob in blobs:
        doc, uri, _ = download_blob(account_url, container_name, credential, blob)
        yield (json.loads(doc), uri)


def dump_list_to_odc(
    account_url,
    container_name,
    credential,
    blob_urls: List[str],
    dc: Datacube,
    stac: Optional[bool] = False,
    update: Optional[bool] = False,
    update_if_exists: Optional[bool] = False,
    allow_unsafe: Optional[bool] = False,
    archive_less_mature: Optional[bool] = False,
    publish_action: Optional[str] = None,
):
    ds_added = 0
    ds_failed = 0
    doc2ds = Doc2Dataset(dc.index)

    for doc, uri in stream_blob_urls(
        account_url, container_name, credential, blob_urls
    ):
        try:
            stac_doc = None
            if stac:
                stac_doc = doc
                doc = stac_transform(doc)
            index_update_dataset(
                doc,
                uri,
                dc,
                doc2ds,
                update=update,
                update_if_exists=update_if_exists,
                allow_unsafe=allow_unsafe,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
                stac_doc=stac_doc,
            )
            ds_added += 1
        except Exception:  # pylint:disable=broad-except
            logging.exception("Failed to add %s", uri)
            ds_failed += 1

    return ds_added, ds_failed


@click.command("azure-to-dc")
@update_flag
@update_if_exists_flag
@allow_unsafe
@transform_stac
@statsd_setting
@archive_less_mature
@publish_action
@click.option(
    "--account_url",
    "-a",
    "account_url",
    default=None,
    help=(
        "If you set the account URL, then you need a different"
        "kind of connection string to if you don't set one"
    ),
)
@click.argument("container_name", type=str, nargs=1)
@click.argument("credential", type=str, nargs=1)
@click.argument("prefix", type=str, nargs=1)
@click.argument("suffix", type=str, nargs=1)
def cli(
    update: bool,
    update_if_exists: bool,
    allow_unsafe: bool,
    stac: bool,
    statsd_setting: str,
    archive_less_mature: bool,
    publish_action: str,
    account_url: str,
    container_name: str,
    credential: str,
    prefix: str,
    suffix: str,
):
    # Set up the datacube first, to ensure we have a connection
    dc = Datacube()
    print(f"Opening AZ Container {container_name} on {account_url}")
    print(f"Searching on prefix '{prefix}' for files matching suffix '{suffix}'")
    yaml_urls = find_blobs(
        container_name, credential, prefix, suffix, account_url=account_url
    )

    # Consume generator and fetch YAML's
    added, failed = dump_list_to_odc(
        account_url,
        container_name,
        credential,
        yaml_urls,
        dc,
        stac=stac,
        update=update,
        update_if_exists=update_if_exists,
        allow_unsafe=allow_unsafe,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
    )

    print(f"Added {added} Datasets, Failed to add {failed} Datasets")
    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:azure_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:azure_to_dc", "action:failed"], statsd_setting
        )


if __name__ == "__main__":
    cli()



import logging
import os
from typing import Iterable, Optional, Union

import click
import pkg_resources
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.utils import changes

from datadog import statsd, initialize

from odc.aws.queue import publish_to_topic
from ._stac import ds_to_stac

ESRI_LANDCOVER_BASE_URI = (
    "https://ai4edataeuwest.blob.core.windows.net/io-lulc/"
    "io-lulc-model-001-v01-composite-v03-supercell-v02-clip-v01/{id}_20200101-20210101.tif"
)

MICROSOFT_PC_STAC_URI = "https://planetarycomputer.microsoft.com/api/stac/v1/"


class IndexingException(Exception):
    """
    Exception to raise for error during SQS to DC indexing/archiving
    """


class SkippedException(Exception):
    """
    Exception to raise for error if dataset exists  and not updating
    """


# A whole bunch of generic Click options
skip_lineage = click.option(
    "--skip-lineage",
    is_flag=True,
    default=False,
    help="Default is not to skip lineage. Set to skip lineage altogether.",
)

fail_on_missing_lineage = click.option(
    "--fail-on-missing-lineage/--auto-add-lineage",
    is_flag=True,
    default=True,
    help=(
        "Default is to fail if lineage documents not present in the database. "
        "Set auto add to try to index lineage documents."
    ),
)

verify_lineage = click.option(
    "--verify-lineage",
    is_flag=True,
    default=False,
    help="Default is no verification. Set to verify parent dataset definitions.",
)

transform_stac = click.option(
    "--stac",
    is_flag=True,
    default=False,
    help="Expect STAC 1.0 metadata and attempt to transform to ODC EO3 metadata.",
)

transform_stac_absolute = click.option(
    "--absolute",
    is_flag=True,
    default=False,
    help="Use absolute paths from the STAC document.",
)

update_flag = click.option(
    "--update",
    is_flag=True,
    default=False,
    help="If set, update instead of add datasets.",
)

update_if_exists_flag = click.option(
    "--update-if-exists",
    is_flag=True,
    default=False,
    help="If the dataset or product already exists, update it instead of skipping it.",
)

allow_unsafe = click.option(
    "--allow-unsafe",
    is_flag=True,
    default=False,
    help="Allow unsafe changes to a dataset. Take care!",
)

skip_check = click.option(
    "--skip-check",
    is_flag=True,
    default=False,
    help="Assume file exists when listing exact file rather than wildcard.",
)

no_sign_request = click.option(
    "--no-sign-request",
    is_flag=True,
    default=False,
    help="Do not sign AWS S3 requests.",
)

request_payer = click.option(
    "--request-payer",
    is_flag=True,
    default=False,
    help="Needed when accessing requester pays public buckets.",
)

archive_less_mature = click.option(
    "--archive-less-mature",
    is_flag=True,
    default=False,
    help=(
        "Archive existing any datasets that match product, "
        "time and region-code, but have lower dataset-maturity."
        "Note: An error will be raised and the dataset add will "
        "fail if a matching dataset with higher or equal dataset-maturity."
    ),
)

publish_action = click.option(
    "--publish-action",
    type=str,
    default=None,
    nargs=1,
    help="SNS topic arn to publish indexing/archiving actions to.",
)

archive = click.option(
    "--archive",
    is_flag=True,
    default=False,
    help="Archive datasets instead of adding them.",
)

limit = click.option(
    "--limit",
    default=None,
    type=int,
    help="Stop indexing after n datasets have been indexed.",
)

bbox = click.option(
    "--bbox",
    type=str,
    default=None,
    help="Comma separated list of bounding box coords, lon-min, lat-min, lon-max, lat-max",
)

statsd_setting = click.option(
    "--statsd-setting",
    is_flag=False,
    default=None,
    help="statsd exporter hostname and port, i.e. prometheus-statsd-exporter:9125",
)


def get_esri_list():
    stream = pkg_resources.resource_stream(__name__, "esri-lc-tiles-list.txt")
    with stream as f:
        for tile in f.readlines():
            tile_id = tile.decode().rstrip("\n")
            yield ESRI_LANDCOVER_BASE_URI.format(id=tile_id)


def index_update_dataset(
    metadata: dict,
    uri: str,
    dc: Datacube,
    doc2ds: Doc2Dataset,
    update: bool = False,
    update_if_exists: bool = False,
    allow_unsafe: bool = False,
    archive_less_mature: Optional[Union[bool, Iterable[str]]] = None,
    publish_action: Optional[str] = None,
    stac_doc: Optional[dict] = None,
) -> int:
    """
    Index and/or update a dataset.  Called by all the **_to_dc CLI tools.

    :param metadata: A dataset metadata dictionary, read from yaml or json, converted from STAC, etc.
    :param uri: The URI of the metadata and associated data.
    :param dc: A datacube object (carries a database index and potentially an active transaction).
    :param doc2ds: A Doc2Dataset object (metadata_type and product resolver)
    :param update: If true, allow update only.
    :param update_if_exists: If true allow insert or update.
    :param allow_unsafe: Allow unsafe (arbitrary) dataset updates.
    :param archive_less_mature: Enforce dataset maturity.
           * If None (the default) or False or an empty iterable, ignore dataset maturity.
           * If True, enforce dataset maturity by looking for existing datasets with same product, region_code and time
             values. If a less mature match is found, it is archived and replaced with the new dataset being inserted.
             If a match of the same or greater maturity is found an IndexException is raised.
           * If an iterable of valid search field names is provided, it is used as the "grouping" fields for
             identifying dataset maturity matches.
             (i.e. `archive_less_mature=True` is the same as `archive_less_mature=['region_code', 'time'])
    :param publish_action: SNS topic arn to publish action to.
    :param stac_doc: STAC document for publication to SNS topic.
    :return: Returns nothing.  Raises an exception if anything goes wrong.
    """
    if uri is None:
        raise IndexingException("Failed to get URI from metadata doc")
    # Make sure we can create a dataset first
    try:
        ds, err = doc2ds(metadata, uri)
    except ValueError as e:
        raise IndexingException(
            f"Exception thrown when trying to create dataset: '{e}'\n The URI was {uri}"
        ) from e
    if ds is None:
        raise IndexingException(
            f"Failed to create dataset with error {err}\n The URI was {uri}"
        )

    if archive_less_mature:
        if archive_less_mature is True:
            # if set explicitly to True, default to [region_code, time]
            archive_less_mature = ["region_code", "time"]
        try:
            dupe_query = {k: getattr(ds.metadata, k) for k in archive_less_mature}
        except AttributeError as e:
            raise IndexingException(
                f"Cannot extract matching value from dataset for maturity check: {e}\n The URI was {uri}"
            )
    else:
        dupe_query = {}

    with dc.index.transaction():
        # Process in a transaction
        archive_ids = []
        archive_stacs = []
        added = False
        updated = False
        if archive_less_mature:
            dupes = dc.index.datasets.search(product=ds.type.name, **dupe_query)
            for dupe in dupes:
                if dupe.id == ds.id:
                    # Same dataset, for update.  Ignore
                    continue
                if dupe.metadata.dataset_maturity <= ds.metadata.dataset_maturity:
                    # Duplicate is as mature, or more mature than ds
                    # E.g. "final" < "nrt"
                    raise IndexingException(
                        f"Matching dataset of maturity {dupe.metadata.dataset_maturity} already exists"
                        "with id: {dupe.id}\n"
                        f"Cannot load dataset of maturity {ds.metadata.dataset_maturity} URI {uri} "
                    )
                archive_ids.append(dupe.id)
                if publish_action:
                    archive_stacs.append(ds_to_stac(dupe))
            if archive_ids:
                dc.index.datasets.archive(archive_ids)

        # Now do something with the dataset
        # Note that any of the exceptions raised below will rollback any archiving performed above.
        if dc.index.datasets.has(metadata.get("id")):
            # Update
            if update or update_if_exists:
                # Set up update fields
                updates = {}
                if allow_unsafe:
                    updates = {tuple(): changes.allow_any}
                # Do the updating
                try:
                    dc.index.datasets.update(ds, updates_allowed=updates)
                    updated = True
                except ValueError as e:
                    raise IndexingException(
                        f"Updating the dataset raised an exception: {e}"
                    )
            else:
                logging.warning("Dataset already exists, not indexing")
                raise SkippedException(
                    f"Dataset {metadata.get('id')} already exists, not indexing"
                )
        else:
            if update:
                # We're expecting to update a dataset, but it doesn't exist
                raise IndexingException(
                    "Can't update dataset because it doesn't exist."
                )
            # Everything is working as expected, add the dataset
            dc.index.datasets.add(ds)
            added = True

    # Transaction committed : Log actions
    for arch_id in archive_ids:
        logging.info("Archived less mature dataset: %s", arch_id)
    if publish_action:
        for arch_stac in archive_stacs:
            publish_to_topic(arn=publish_action, action="ARCHIVED", stac=arch_stac)

    if added:
        logging.info("New Dataset Added: %s", ds.id)
        if publish_action:
            # if STAC was not provided, generate from dataset
            stac_doc = stac_doc if stac_doc else ds_to_stac(ds)
            publish_to_topic(arn=publish_action, action="ADDED", stac=stac_doc)

    if updated:
        logging.info("Existing Dataset Updated: %s", ds.id)


def statsd_gauge_reporting(value, tags=None, statsd_setting="localhost:8125"):
    if tags is None:
        tags = []
    host = statsd_setting.split(":")[0]
    port = statsd_setting.split(":")[1]
    options = {"statsd_host": host, "statsd_port": port}
    initialize(**options)

    if os.environ.get("HOSTNAME"):
        tags.append(f"pod:{os.getenv('HOSTNAME')}")
    statsd.gauge("datacube_index", value, tags=tags)



#!/usr/bin/env python3
"""
Index the ESA Worldcover data automatically.
"""
import concurrent.futures
import logging
import sys
from math import ceil, floor
from typing import Tuple

import click
import pystac
import rasterio
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.utils import read_documents
from odc.apps.dc_tools.utils import (
    bbox,
    index_update_dataset,
    limit,
    update_if_exists_flag,
    archive_less_mature,
    statsd_gauge_reporting,
    statsd_setting,
    publish_action,
)
from rio_stac import create_stac_item

from ._stac import stac_transform

PRODUCT = (
    "https://raw.githubusercontent.com/opendatacube/"
    "datacube-dataset-config/master/products/esa_worldcover.odc-product.yaml"
)

# URIs need north/south, which is N00 and east/west, which is E000
URI_TEMPLATE = (
    "https://esa-worldcover.s3.eu-central-1.amazonaws.com/"
    "v100/2020/map/ESA_WorldCover_10m_2020_v100_{ns}{ew}_Map.tif"
)


def _unpack_bbox(bounding_box: str) -> Tuple[int, int, int, int]:
    """Set up a left, bottom, right, top bbox with 3 degree offset"""
    left = floor(bounding_box[0])
    bottom = floor(bounding_box[1])
    right = ceil(bounding_box[2])
    top = ceil(bounding_box[3])

    left_offset = left % 3
    bottom_offset = bottom % 3
    right_offset = right % 3
    top_offset = top % 3

    if right_offset != 0:
        right_offset -= 3
    if top_offset != 0:
        top_offset -= 3

    left = left - left_offset
    bottom = bottom - bottom_offset
    right = right - right_offset
    top = top - top_offset
    return left, bottom, right, top


def add_odc_product(dc: Datacube):
    for _, doc in read_documents(PRODUCT):
        dc.index.products.add_document(doc)
    print("Product definition added")


def get_tile_uris(bounding_box: str) -> Tuple[str, str]:
    # Validate the bounding_box
    if bounding_box is None:
        logging.warning(
            "No BBOX provided, running full extent... this will take a long time."
        )
        bounding_box = (-180, -90, 180, 90)
    else:
        bounding_box = bounding_box.split(",")
        if len(bounding_box) != 4:
            raise ValueError("bounding_box must be in the format: minx,miny,maxx,maxy")
        bounding_box = [float(x) for x in bounding_box]

    # The tiles are 3 x 3 degree, starting at 0,3...
    # so we need some logic to get to those numbers
    left, bottom, right, top = _unpack_bbox(bounding_box)

    x_range = range(left, right, 3)
    y_range = range(bottom, top, 3)

    for x in x_range:
        for y in y_range:
            if x < 0:
                x_str = f"W{abs(x):03d}"
            else:
                x_str = f"E{x:03d}"
            if y < 0:
                y_str = f"S{abs(y):02d}"
            else:
                y_str = f"N{y:02d}"
            yield (
                URI_TEMPLATE.format(ns=y_str, ew=x_str),
                f"{x_str}_{y_str}",
            )


def process_uri_tile(
    uri_tile: Tuple[str, str, str],
    dc: Datacube,
    doc2ds: Doc2Dataset,
    update_if_exists: bool = True,
    archive_less_mature: bool = False,
    publish_action: str = None,
) -> Tuple[pystac.Item, str]:
    product_name = "esa_worldcover"
    uri, tile = uri_tile
    properties = {
        "odc:product": product_name,
        "odc:region_code": tile,
        "start_datetime": "2020-01-01",
        "end_datetime": "2020-12-31",
    }

    with rasterio.Env(aws_unsigned=True, GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR"):
        item = create_stac_item(
            uri,
            collection=product_name,
            with_proj=True,
            properties=properties,
            asset_media_type=pystac.MediaType.COG,
            asset_name="classification",
        )

    index_update_dataset(
        stac_transform(item.to_dict()),
        uri,
        dc,
        doc2ds,
        update_if_exists=update_if_exists,
        allow_unsafe=True,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        stac_doc=item.to_dict(),
    )

    return True


def esa_wc_to_dc(
    dc: Datacube,
    bounding_box,
    limit: int,
    update: bool,
    n_workers: int = 100,
    archive_less_mature: bool = False,
    publish_action: str = None,
) -> Tuple[int, int]:
    doc2ds = Doc2Dataset(dc.index)

    # Get a generator of (uris)
    uris_tiles = list(get_tile_uris(bounding_box))
    if limit:
        uris_tiles = uris_tiles[0:limit]

    # Do the indexing of all the things
    success = 0
    failure = 0

    sys.stdout.write(f"Starting ESA Worldcover indexing with {n_workers} workers...\n")

    with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
        future_to_uri = {
            executor.submit(
                process_uri_tile,
                uri_tile,
                dc,
                doc2ds,
                update_if_exists=update,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
            ): uri_tile[0]
            for uri_tile in uris_tiles
        }
        for future in concurrent.futures.as_completed(future_to_uri):
            uri = future_to_uri[future]
            try:
                _ = future.result()
                success += 1
                if success % 10 == 0:
                    sys.stdout.write(f"\rAdded {success} datasets...")
            except rasterio.errors.RasterioIOError:
                logging.info("Couldn't read file %s", uri, exc_info=True)
            except Exception:  # pylint:disable=broad-except
                logging.exception("Failed to handle uri %s", uri)
                failure += 1
    sys.stdout.write("\r")

    return success, failure


@click.command("esa-wc-to-dc")
@limit
@update_if_exists_flag
@bbox
@click.option(
    "--add-product",
    is_flag=True,
    default=False,
    help="If set, add the product too",
)
@archive_less_mature
@publish_action
@click.option(
    "--workers",
    default=20,
    type=int,
    help="Number of threads to use to process, default 20",
)
@statsd_setting
def cli(
    limit,
    update_if_exists,
    bbox,
    add_product,
    archive_less_mature,
    workers,
    statsd_setting,
    publish_action,
):
    """
    Index the ESA WorldCover product automatically.
    """

    dc = Datacube()

    if add_product:
        add_odc_product(dc)

    print(f"Indexing ESA WorldCover with bounding box of {bbox}")

    added, failed = esa_wc_to_dc(
        dc,
        bbox,
        limit,
        update_if_exists,
        n_workers=workers,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
    )

    print(f"Added {added} Datasets, failed {failed} Datasets")

    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:esa_worldcover_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:esa_worldcover_to_dc", "action:failed"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()
